# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root
# for license information.

"""pytest configuration.
"""

pytest_plugins = ["tests.pytest_fixtures", "tests.pytest_hooks"]
